from .processing import \
    map, \
    filter, \
    split, \
    expand, \
    combine, \
    join

from .transformer import \
    Mapper, \
    Filter, \
    Splitter, \
    Expander, \
    Combiner, \
    Joiner, \
    Manipulator, \
    Window, \
    Flat, \
    Union, \
    Pipeline

from .manipulation import \
    windowed, \
    flatten



import daproli as dp

import unittest


class ManipulationTest(unittest.TestCase):

    def test_windowed(self):
        data = range(10)
        res = dp.windowed(data, 2, step=2)

        self.assertEqual([[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]], res)

    def test_flatten(self):
        data = [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
        res = dp.flatten(data)

        self.assertEqual([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], res)


from itertools import product
from joblib import Parallel, delayed
from tqdm import tqdm

from .utils import _get_return_type, _apply_func


def map(func, data, expand_args=True, n_jobs=1, verbose=0, **kwargs):
    '''
    dp.map applies a transformation function to a collection of data items.

    Parameters
    -----------
    :param func: the mapping function
    :param data: an iterable collection of data
    :param expand_args: true if args should be expanded, false otherwise
    :param n_jobs: amount of used threads/processes
    :param verbose: verbosity level for tqdm / joblib
    :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
    :return: the transformed data list

    Examples
    -----------
    >>> import daproli as dp
    >>> names = ['John', 'Susan', 'Mike']
    >>> dp.map(lambda n : n.lower(), names)
    ['john', 'susan', 'mike']
    '''
    ret_type = _get_return_type(data)
    func_ = lambda args : _apply_func(func, args, expand_args)

    if n_jobs == 1:
        return ret_type([func_(item) for item in tqdm(data, disable=verbose < 1)])

    return ret_type(Parallel(n_jobs=n_jobs, verbose=verbose, **kwargs)(delayed(func_)(item) for item in data))


def filter(pred, data, expand_args=True, n_jobs=1, verbose=0, **kwargs):
    '''
    dp.filter applies a filter predicate to a collection of data items.

    Parameters
    -----------
    :param pred: the filter predicate
    :param data: an iterable collection of data
    :param expand_args: true if args should be expanded, false otherwise
    :param n_jobs: amount of used threads/processes
    :param verbose: verbosity level for tqdm / joblib
    :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
    :return: the filtered data list

    Examples
    -----------
    >>> import daproli as dp
    >>> names = ['John', 'Susan', 'Mike']
    >>> dp.filter(lambda n : len(n) % 2 == 0, names)
    ['John', 'Mike']
    '''
    ret_type = _get_return_type(data)
    pred_ = lambda args: _apply_func(pred, args, expand_args)

    if n_jobs == 1:
        return ret_type([item for item in tqdm(data, disable=verbose < 1) if pred_(item)])

    labels = Parallel(n_jobs=n_jobs, verbose=verbose, **kwargs)(delayed(pred_)(item) for item in data)
    return ret_type([item for item, label in zip(data, labels) if label])


def split(func, data, return_labels=False, expand_args=True, n_jobs=1, verbose=0, **kwargs):
    '''
    dp.split applies a discriminator function to a collection of data items.

    Parameters
    -----------
    :param func: the discriminator function
    :param data: an iterable collection of data
    :param return_labels: true if the associated labels should be returned, false otherwise
    :param expand_args: true if args should be expanded, false otherwise
    :param n_jobs: amount of used threads/processes
    :param verbose: verbosity level for tqdm / joblib
    :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
    :return: the transformed data lists

    Examples
    -----------
    >>> import daproli as dp
    >>> numbers = range(10)
    >>> dp.split(lambda x : x % 2 == 0, numbers)
    [[1, 3, 5, 7, 9], [0, 2, 4, 6, 8]]
    '''
    ret_type = _get_return_type(data)
    func_ = lambda args: _apply_func(func, args, expand_args)

    if n_jobs == 1:
        labels = [func_(item) for item in tqdm(data, disable=verbose < 1)]
    else:
        labels = Parallel(n_jobs=n_jobs, verbose=verbose, **kwargs)(delayed(func_)(item) for item in data)

    container = {label : list() for label in set(labels)}
    for item, label in zip(data, labels): container[label].append(item)

    if return_labels:
        return [(label, ret_type(container[label])) for label in sorted(container)]

    return [ret_type(container[label]) for label in sorted(container)]


def expand(func, data, expand_args=True, n_jobs=1, verbose=0, **kwargs):
    '''
    dp.expand applies an expansion function to a collection of data items.

    Parameters
    -----------
    :param func: the expansion function
    :param data: an iterable collection of data
    :param expand_args: true if args should be expanded, false otherwise
    :param n_jobs: amount of used threads/processes
    :param verbose: verbosity level for tqdm / joblib
    :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
    :return: the transformed data lists

    Examples
    -----------
    >>> import daproli as dp
    >>> numbers = range(10)
    >>> dp.expand(lambda x : (x, x**2), numbers)
    [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]]
    '''
    ret_type = _get_return_type(data)
    func_ = lambda args: _apply_func(func, args, expand_args)

    if n_jobs == 1:
        expanded = [func_(item) for item in tqdm(data, disable=verbose < 1)]
    else:
        expanded = Parallel(n_jobs=n_jobs, verbose=verbose, **kwargs)(delayed(func_)(item) for item in data)

    if len(expanded) == 0:
        return ret_type(expanded)

    assert all([len(expanded[idx]) == len(expanded[idx+1]) for idx in range(len(expanded)-1)])
    container = [list() for _ in range(len(expanded[0]))]

    for items in expanded:
        for idx, item in enumerate(items): container[idx].append(item)

    return [ret_type(items) for items in container]


def combine(func, *data, expand_args=True, n_jobs=1, verbose=0, **kwargs):
    '''
    dp.combine applies a combination function to multiple collections of data items.

    Parameters
    -----------
    :param func: the combination function
    :param data: iterable collections of data
    :param expand_args: true if args should be expanded, false otherwise
    :param n_jobs: amount of used threads/processes
    :param verbose: verbosity level for tqdm / joblib
    :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
    :return: the combined data list

    Examples
    -----------
    >>> import daproli as dp
    >>> even_numbers = range(0, 10, 2)
    >>> odd_numbers = range(1, 10, 2)
    >>> dp.combine(lambda x, y : (x,y), even_numbers, odd_numbers)
    [(0, 1), (2, 3), (4, 5), (6, 7), (8, 9)]
    '''
    func_ = lambda args: _apply_func(func, args, expand_args)

    if n_jobs == 1:
        return [func_(items) for items in tqdm(zip(*data), disable=verbose < 1)]

    return Parallel(n_jobs=n_jobs, verbose=verbose, **kwargs)(delayed(func_)(items) for items in zip(*data))


def join(pred, *data, expand_args=True, n_jobs=1, verbose=0, **kwargs):
    '''
    dp.join applies a join predicate to multiple collections of data items.

    Parameters
    -----------
    :param pred: the join predicate
    :param data: iterable collections of data
    :param expand_args: true if args should be expanded, false otherwise
    :param n_jobs: amount of used threads/processes
    :param verbose: verbosity level for tqdm / joblib
    :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
    :return: the joined data list

    Examples
    -----------
    >>> import daproli as dp
    >>> even_numbers = range(0, 10, 2)
    >>> odd_numbers = range(1, 10, 2)
    >>> dp.join(lambda x, y : y-x == 3, even_numbers, odd_numbers)
    [(0, 3), (2, 5), (4, 7), (6, 9)]
    '''
    pred_ = lambda args: _apply_func(pred, args, expand_args)

    if n_jobs == 1:
        return [items for items in tqdm(product(*data), disable=verbose < 1) if pred_(items)]

    labels = Parallel(n_jobs=n_jobs, verbose=verbose, **kwargs)(delayed(pred_)(items) for items in product(*data))
    return [items for items, label in zip(product(*data), labels) if label]




from collections.abc import Iterable

import numpy as np


def _get_return_type(data):
    '''
    An utility function in order to determine the correct return type
    for the transformation functions.

    Parameters
    -----------
    :param data: an iterable collection of data
    :return: the return type
    '''
    if isinstance(data, range):
        return list

    if isinstance(data, zip):
        return list

    if isinstance(data, enumerate):
        return list

    if isinstance(data, np.ndarray):
        return np.array

    return type(data)


def _apply_func(func, args, expand_args):
    '''
    An utility function to apply a given function with either
    zipped or unzipped args.

    Parameters
    -----------
    :param func: the function to apply
    :param args: the args to apply
    :param expand_args: true if args should be expanded, false otherwise
    :return: the function result
    '''
    if expand_args is True and isinstance(args, Iterable) and not isinstance(args, str):
        return func(*args)

    return func(args)


from .utils import _get_return_type


def windowed(data, size, step=1):
    '''
    dp.windowed applies a window function to a collection of data items.

    Parameters
    -----------
    :param data: an iterable collection of data
    :param size: the window size
    :param step: the window step
    :return: the windowed data list

    Examples
    -----------
    >>> import daproli as dp
    >>> numbers = range(10)
    >>> dp.windowed(numbers, 2, step=2)
    [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
    '''
    ret_type = _get_return_type(data)
    return [ret_type(data[i:i+size]) for i in range(0, len(data)-(size-1), step)]


def flatten(data):
    '''
    dp.flatten applies a flatten function to a collection of data items.

    Parameters
    -----------
    :param data: an iterable collection of data
    :return: the flattened data list

    Examples
    -----------
    >>> import daproli as dp
    >>> dp.flatten([[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]])
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    '''
    ret_type = _get_return_type(data)
    return ret_type([item for sub in data for item in sub])



from joblib import Parallel, delayed
from tqdm import tqdm

from .processing import map, filter, split, expand, combine, join
from .manipulation import windowed, flatten


class BaseTransformer:
    '''
    The BaseTransformer defines a generic data transformation pattern that
    can be implemented with a number of data processing concepts.
    '''
    def transform(self, data, *args, **kwargs):
        raise NotImplementedError()


class Mapper(BaseTransformer):

    def __init__(self, func, expand_args=True, n_jobs=1, verbose=0, **kwargs):
        '''
        dp.Mapper is the respective transformer for dp.map.

        Parameters
        -----------
        :param func: the mapping function
        :param expand_args: true if args should be expanded, false otherwise
        :param n_jobs: amount of used threads/processes
        :param verbose: verbosity level for tqdm / joblib
        :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
        '''
        self.func = func
        self.expand_args = expand_args
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.kwargs = kwargs

    def transform(self, data, *args, **kwargs):
        return map(self.func, data, expand_args=self.expand_args, n_jobs=self.n_jobs, verbose=self.verbose, **self.kwargs)


class Filter(BaseTransformer):

    def __init__(self, pred, expand_args=True, n_jobs=1, verbose=0, **kwargs):
        '''
        dp.Filter is the respective transformer for dp.filter.

        Parameters
        -----------
        :param pred: the filter predicate
        :param expand_args: true if args should be expanded, false otherwise
        :param n_jobs: amount of used threads/processes
        :param verbose: verbosity level for tqdm / joblib
        :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
        '''
        self.pred = pred
        self.expand_args = expand_args
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.kwargs = kwargs

    def transform(self, data, *args, **kwargs):
        return filter(self.pred, data, expand_args=self.expand_args, n_jobs=self.n_jobs, verbose=self.verbose, **self.kwargs)


class Splitter(BaseTransformer):

    def __init__(self, func, return_labels=False, expand_args=True, n_jobs=1, verbose=0, **kwargs):
        '''
        dp.Splitter is the respective transformer for dp.split.

        Parameters
        -----------
        :param func: the discriminator function
        :param return_labels: true if the associated labels should be returned, false otherwise
        :param expand_args: true if args should be expanded, false otherwise
        :param n_jobs: amount of used threads/processes
        :param verbose: verbosity level for tqdm / joblib
        :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
        '''
        self.func = func
        self.return_labels = return_labels
        self.expand_args = expand_args
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.kwargs = kwargs

    def transform(self, data, *args, **kwargs):
        return split(self.func, data, return_labels=self.return_labels, expand_args=self.expand_args, n_jobs=self.n_jobs,
                     verbose=self.verbose, **self.kwargs)


class Expander(BaseTransformer):

    def __init__(self, func, expand_args=True, n_jobs=1, verbose=0, **kwargs):
        '''
        dp.Expander is the respective transformer for dp.expand.

        Parameters
        -----------
        :param func: the expansion function
        :param expand_args: true if args should be expanded, false otherwise
        :param n_jobs: amount of used threads/processes
        :param verbose: verbosity level for tqdm / joblib
        :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
        '''
        self.func = func
        self.expand_args = expand_args
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.kwargs = kwargs

    def transform(self, data, *args, **kwargs):
        return expand(self.func, data, expand_args=self.expand_args, n_jons=self.n_jobs, verbose=self.verbose, **self.kwargs)


class Combiner(BaseTransformer):

    def __init__(self, func, expand_args=True, n_jobs=1, verbose=0, **kwargs):
        '''
        dp.Combiner is the respective transformer for dp.combine.

        Parameters
        -----------
        :param func: the combination function
        :param expand_args: true if args should be expanded, false otherwise
        :param n_jobs: amount of used threads/processes
        :param verbose: verbosity level for tqdm / joblib
        :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
        '''
        self.func = func
        self.expand_args = expand_args
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.kwargs = kwargs

    def transform(self, data, *args, **kwargs):
        return combine(self.func, *data, expand_args=self.expand_args, n_jobs=self.n_jobs, verbose=self.verbose, **self.kwargs)


class Joiner(BaseTransformer):

    def __init__(self, func, expand_args=True, n_jobs=1, verbose=0, **kwargs):
        '''
        dp.Joiner is the respective transformer for dp.join.

        Parameters
        -----------
        :param func: the join function
        :param expand_args: true if args should be expanded, false otherwise
        :param n_jobs: amount of used threads/processes
        :param verbose: verbosity level for tqdm / joblib
        :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
        '''
        self.func = func
        self.expand_args = expand_args
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.kwargs = kwargs

    def transform(self, data, *args, **kwargs):
        return join(self.func, *data, expand_args=self.expand_args, n_jobs=self.n_jobs, verbose=self.verbose, **self.kwargs)


class Manipulator(BaseTransformer):

    def __init__(self, func, *args, **kwargs):
        '''
        dp.Manipulator is a transformer to manipulate the entire collection of data items.

        Parameters
        -----------
        :param func: the manipulation function
        :param args: additional args for func
        :param kwargs: additional kwargs for func
        '''
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def transform(self, data, *args, **kwargs):
        return self.func(data, *self.args, **self.kwargs)


class Window(BaseTransformer):

    def __init__(self, size, step=1):
        '''
        dp.Window is the respective transformer for dp.windowed.

        Parameters
        -----------
        :param data: an iterable collection of data
        :param size: the window size
        :param step: the window step
        '''
        self.size = size
        self.step = step

    def transform(self, data, *args, **kwargs):
        return windowed(data, self.size, step=self.step)


class Flat(BaseTransformer):

    def __init__(self):
        '''
        dp.Flat is the respective transformer for dp.flatten.
        '''
        pass

    def transform(self, data, *args, **kwargs):
        return flatten(data)


class Union(BaseTransformer):

    def __init__(self, *transformers, n_jobs=1, verbose=0, **kwargs):
        '''
        dp.Union is a construct to manipulate mutli-collections of data tiems.

        Parameters
        -----------
        :param transformers: the transformers for the respective collections of data items
        :param n_jobs: amount of used threads/processes
        :param verbose: verbosity level for tqdm / joblib
        :param kwargs: additional arguments for joblib.Parallel, e.g. backend='loky'
        '''
        self.transformers = transformers
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.kwargs = kwargs

    def transform(self, data, *args, **kwargs):
        if self.n_jobs == 1:
            return [transformer.transform(items, *args, **kwargs)
                    for transformer, items in tqdm(zip(self.transformers, data), disable=self.verbose < 1)]

        return Parallel(n_jobs=self.n_jobs, verbose=self.verbose, **self.kwargs)(delayed(transformer.transform)
                (items, *args, **kwargs) for transformer, items in zip(self.transformers, data))


class Pipeline(BaseTransformer):

    def __init__(self, *transformers, verbose=0):
        '''
        dp.Pipeline is a construct to pipe a collection of transformers.

        Parameters
        -----------
        :param transformers: the transformer sequence to apply
        :param verbose: verbosity level for tqdm
        '''
        self.transformers = list(transformers)
        self.verbose = verbose

    def transform(self, data, *args, **kwargs):
        res = data

        for transformer in tqdm(self.transformers, disable=self.verbose < 1):
            res = transformer.transform(res, *args, **kwargs)

        return res
